@@ -1,6 +1,6 @@
 //
 //  ========================================================================
-//  Copyright (c) 1995-2020 Mort Bay Consulting Pty Ltd and others.
+//  Copyright (c) 1995-2022 Mort Bay Consulting Pty Ltd and others.
 //  ------------------------------------------------------------------------
 //  All rights reserved. This program and the accompanying materials
 //  are made available under the terms of the Eclipse Public License v1.0
@@ -29,7 +29,9 @@ import java.net.URL;
 import java.net.URLConnection;
 import java.nio.channels.FileChannel;
 import java.nio.channels.ReadableByteChannel;
+import java.nio.file.Files;
 import java.nio.file.InvalidPathException;
+import java.nio.file.Path;
 import java.nio.file.StandardOpenOption;
 import java.security.Permission;
 
@@ -39,26 +41,25 @@ import org.eclipse.jetty.util.URIUtil;
 import org.eclipse.jetty.util.log.Log;
 import org.eclipse.jetty.util.log.Logger;
 
-
-/* ------------------------------------------------------------ */
-/** File Resource.
+/**
+ * File Resource.
  *
  * Handle resources of implied or explicit file type.
  * This class can check for aliasing in the filesystem (eg case
- * insensitivity).  By default this is turned on, or it can be controlled 
+ * insensitivity).  By default this is turned on, or it can be controlled
  * by calling the static method @see FileResource#setCheckAliases(boolean)
- * 
+ *
+ * @deprecated Use {@link PathResource}
  */
+@Deprecated
 public class FileResource extends Resource
 {
     private static final Logger LOG = Log.getLogger(FileResource.class);
 
-    /* ------------------------------------------------------------ */
     private final File _file;
-    private final String _uri;
+    private final URI _uri;
     private final URI _alias;
-    
-    /* -------------------------------------------------------- */
+
     public FileResource(URL url)
         throws IOException, URISyntaxException
     {
@@ -66,10 +67,10 @@ public class FileResource extends Resource
         try
         {
             // Try standard API to convert URL to file.
-            file =new File(url.toURI());
+            file = new File(url.toURI());
             assertValidPath(file.toString());
         }
-        catch (URISyntaxException e) 
+        catch (URISyntaxException e)
         {
             throw e;
         }
@@ -77,143 +78,208 @@ public class FileResource extends Resource
         {
             if (!url.toString().startsWith("file:"))
                 throw new IllegalArgumentException("!file:");
-            
+
             LOG.ignore(e);
             try
             {
                 // Assume that File.toURL produced unencoded chars. So try encoding them.
-                String file_url="file:"+URIUtil.encodePath(url.toString().substring(5));           
-                URI uri = new URI(file_url);
-                if (uri.getAuthority()==null) 
+                String fileUrl = "file:" + URIUtil.encodePath(url.toString().substring(5));
+                URI uri = new URI(fileUrl);
+                if (uri.getAuthority() == null)
                     file = new File(uri);
                 else
-                    file = new File("//"+uri.getAuthority()+URIUtil.decodePath(url.getFile()));
+                    file = new File("//" + uri.getAuthority() + URIUtil.decodePath(url.getFile()));
             }
-            catch (Exception e2)
+            catch (Exception ex2)
             {
-                LOG.ignore(e2);
+                LOG.ignore(ex2);
                 // Still can't get the file.  Doh! try good old hack!
-                URLConnection connection=url.openConnection();
+                URLConnection connection = url.openConnection();
                 Permission perm = connection.getPermission();
-                file = new File(perm==null?url.getFile():perm.getName());
+                file = new File(perm == null ? url.getFile() : perm.getName());
             }
         }
-        
-        _file=file;
-        _uri=normalizeURI(_file,url.toURI());
-        _alias=checkFileAlias(_file);
+
+        _file = file;
+        _uri = normalizeURI(_file, url.toURI());
+        _alias = checkFileAlias(_uri, _file);
     }
 
-    /* -------------------------------------------------------- */
     public FileResource(URI uri)
     {
-        File file=new File(uri);
-        _file=file;
-        URI file_uri=_file.toURI();
-        _uri=normalizeURI(_file,uri);
-        assertValidPath(file.toString());
-
-        // Is it a URI alias?
-        if (!URIUtil.equalsIgnoreEncodings(_uri,file_uri.toString()))
-            _alias=_file.toURI();
-        else
-            _alias=checkFileAlias(_file);
-    }
-
-    /* -------------------------------------------------------- */
-    FileResource(File file)
-    {
-        assertValidPath(file.toString());
-        _file=file;
-        _uri=normalizeURI(_file,_file.toURI());
-        _alias=checkFileAlias(_file);
-    }
-
-    /* -------------------------------------------------------- */
-    private static String normalizeURI(File file, URI uri)
-    {
-        String u =uri.toASCIIString();
-        if (file.isDirectory())
+        File file = new File(uri);
+        _file = file;
+        try
         {
-            if(!u.endsWith("/"))
-                u+="/";
-        } 
-        else if (file.exists() && u.endsWith("/"))
-            u=u.substring(0,u.length()-1);
-        return u;
+            URI fileUri = _file.toURI();
+            _uri = normalizeURI(_file, uri);
+            assertValidPath(file.toString());
+
+            // Is it a URI alias?
+            if (!URIUtil.equalsIgnoreEncodings(_uri.toASCIIString(), fileUri.toString()))
+                _alias = _file.toURI();
+            else
+                _alias = checkFileAlias(_uri, _file);
+        }
+        catch (URISyntaxException e)
+        {
+            throw new InvalidPathException(_file.toString(), e.getMessage())
+            {
+                {
+                    initCause(e);
+                }
+            };
+        }
     }
 
-    /* -------------------------------------------------------- */
-    private static URI checkFileAlias(File file)
+    public FileResource(File file)
+    {
+        assertValidPath(file.toString());
+        _file = file;
+        try
+        {
+            _uri = normalizeURI(_file, _file.toURI());
+        }
+        catch (URISyntaxException e)
+        {
+            throw new InvalidPathException(_file.toString(), e.getMessage())
+            {
+                {
+                    initCause(e);
+                }
+            };
+        }
+        _alias = checkFileAlias(_uri, _file);
+    }
+
+    public FileResource(File base, String childPath)
+    {
+        String encoded = URIUtil.encodePath(childPath);
+
+        _file = new File(base, childPath);
+
+        // The encoded path should be a suffix of the resource (give or take a directory / )
+        URI uri;
+        try
+        {
+            if (base.isDirectory())
+            {
+                // treat all paths being added as relative
+                uri = new URI(URIUtil.addEncodedPaths(base.toURI().toASCIIString(), encoded));
+            }
+            else
+            {
+                uri = new URI(base.toURI().toASCIIString() + encoded);
+            }
+        }
+        catch (final URISyntaxException e)
+        {
+            throw new InvalidPathException(base.toString() + childPath, e.getMessage())
+            {
+                {
+                    initCause(e);
+                }
+            };
+        }
+
+        _uri = uri;
+        _alias = checkFileAlias(_uri, _file);
+    }
+
+    @Override
+    public boolean isSame(Resource resource)
     {
         try
         {
-            String abs=file.getAbsolutePath();
-            String can=file.getCanonicalPath();
+            if (resource instanceof PathResource)
+            {
+                Path path = ((PathResource)resource).getPath();
+                return Files.isSameFile(getFile().toPath(), path);
+            }
+            if (resource instanceof FileResource)
+            {
+                Path path = ((FileResource)resource).getFile().toPath();
+                return Files.isSameFile(getFile().toPath(), path);
+            }
+        }
+        catch (IOException e)
+        {
+            if (LOG.isDebugEnabled())
+                LOG.debug("ignored", e);
+        }
+        return false;
+    }
+
+    private static URI normalizeURI(File file, URI uri) throws URISyntaxException
+    {
+        String u = uri.toASCIIString();
+        if (file.isDirectory())
+        {
+            if (!u.endsWith("/"))
+                u += "/";
+        }
+        else if (file.exists() && u.endsWith("/"))
+            u = u.substring(0, u.length() - 1);
+        return new URI(u);
+    }
+
+    private static URI checkFileAlias(final URI uri, final File file)
+    {
+        try
+        {
+            if (!URIUtil.equalsIgnoreEncodings(uri, file.toURI()))
+            {
+                // Return alias pointing to Java File normalized URI
+                return new File(uri).getAbsoluteFile().toURI();
+            }
+
+            String abs = file.getAbsolutePath();
+            String can = file.getCanonicalPath();
 
             if (!abs.equals(can))
             {
                 if (LOG.isDebugEnabled())
-                    LOG.debug("ALIAS abs={} can={}",abs,can);
+                    LOG.debug("ALIAS abs={} can={}", abs, can);
 
-                URI alias=new File(can).toURI();
+                URI alias = new File(can).toURI();
                 // Have to encode the path as File.toURI does not!
-                return new URI("file://"+URIUtil.encodePath(alias.getPath()));  
+                return new URI("file://" + URIUtil.encodePath(alias.getPath()));
             }
         }
-        catch(Exception e)
+        catch (Exception e)
         {
-            LOG.warn("bad alias for {}: {}",file,e.toString());
+            LOG.warn("bad alias for {}: {}", file, e.toString());
             LOG.debug(e);
             try
             {
-                return new URI("http://eclipse.org/bad/canonical/alias");
+                return new URI("https://eclipse.org/bad/canonical/alias");
             }
-            catch(Exception e2)
+            catch (Exception ex2)
             {
-                LOG.ignore(e2);
+                LOG.ignore(ex2);
                 throw new RuntimeException(e);
             }
         }
 
         return null;
     }
-    
-    /* -------------------------------------------------------- */
+
     @Override
     public Resource addPath(String path)
-        throws IOException,MalformedURLException
+        throws IOException
     {
         assertValidPath(path);
-        path = org.eclipse.jetty.util.URIUtil.canonicalPath(path);
 
-        if (path==null)
-            throw new MalformedURLException();   
-        
+        // Check that the path is within the root,
+        // but use the original path to create the
+        // resource, to preserve aliasing.
+        if (URIUtil.canonicalPath(path) == null)
+            throw new MalformedURLException(path);
+
         if ("/".equals(path))
             return this;
-        
-        path=URIUtil.encodePath(path);
-        // The encoded path should be a suffix of the resource (give or take a directory / )
-        URI uri;
-        try
-        {
-            if (_file.isDirectory())
-            {
-                // treat all paths being added as relative
-                uri=new URI(URIUtil.addPaths(_uri,path));
-            }
-            else
-            {
-                uri=new URI(_uri+path);
-            }
-        }
-        catch(final URISyntaxException e)
-        {
-            throw new InvalidPathException(path, e.getMessage());
-        }
 
-        return new FileResource(uri);
+        return new FileResource(_file, path);
     }
 
     private void assertValidPath(String path)
@@ -225,14 +291,12 @@ public class FileResource extends Resource
         }
     }
 
-    /* ------------------------------------------------------------ */
     @Override
     public URI getAlias()
     {
         return _alias;
     }
-    
-    /* -------------------------------------------------------- */
+
     /**
      * Returns true if the resource exists.
      */
@@ -241,8 +305,7 @@ public class FileResource extends Resource
     {
         return _file.exists();
     }
-        
-    /* -------------------------------------------------------- */
+
     /**
      * Returns the last modified time
      */
@@ -252,17 +315,15 @@ public class FileResource extends Resource
         return _file.lastModified();
     }
 
-    /* -------------------------------------------------------- */
     /**
      * Returns true if the resource is a container/directory.
      */
     @Override
     public boolean isDirectory()
     {
-        return _file.exists() && _file.isDirectory() || _uri.endsWith("/");
+        return _file.exists() && _file.isDirectory() || _uri.toASCIIString().endsWith("/");
     }
 
-    /* --------------------------------------------------------- */
     /**
      * Return the length of the resource
      */
@@ -271,9 +332,7 @@ public class FileResource extends Resource
     {
         return _file.length();
     }
-        
 
-    /* --------------------------------------------------------- */
     /**
      * Returns the name of the resource
      */
@@ -282,8 +341,7 @@ public class FileResource extends Resource
     {
         return _file.getAbsolutePath();
     }
-        
-    /* ------------------------------------------------------------ */
+
     /**
      * Returns an File representing the given resource or NULL if this
      * is not possible.
@@ -293,8 +351,7 @@ public class FileResource extends Resource
     {
         return _file;
     }
-        
-    /* --------------------------------------------------------- */
+
     /**
      * Returns an input stream to the resource
      */
@@ -304,14 +361,12 @@ public class FileResource extends Resource
         return new FileInputStream(_file);
     }
 
-    /* ------------------------------------------------------------ */
     @Override
     public ReadableByteChannel getReadableByteChannel() throws IOException
     {
-        return FileChannel.open(_file.toPath(),StandardOpenOption.READ);
+        return FileChannel.open(_file.toPath(), StandardOpenOption.READ);
     }
-        
-    /* --------------------------------------------------------- */
+
     /**
      * Deletes the given resource
      */
@@ -322,81 +377,77 @@ public class FileResource extends Resource
         return _file.delete();
     }
 
-    /* --------------------------------------------------------- */
     /**
      * Rename the given resource
      */
     @Override
-    public boolean renameTo( Resource dest)
+    public boolean renameTo(Resource dest)
         throws SecurityException
     {
-        if( dest instanceof FileResource)
-            return _file.renameTo( ((FileResource)dest)._file);
+        if (dest instanceof FileResource)
+            return _file.renameTo(((FileResource)dest)._file);
         else
             return false;
     }
 
-    /* --------------------------------------------------------- */
     /**
      * Returns a list of resources contained in the given resource
      */
     @Override
     public String[] list()
     {
-        String[] list =_file.list();
-        if (list==null)
+        String[] list = _file.list();
+        if (list == null)
             return null;
-        for (int i=list.length;i-->0;)
+        for (int i = list.length; i-- > 0; )
         {
-            if (new File(_file,list[i]).isDirectory() &&
+            if (new File(_file, list[i]).isDirectory() &&
                 !list[i].endsWith("/"))
-                list[i]+="/";
+                list[i] += "/";
         }
         return list;
     }
-    
-    /* ------------------------------------------------------------ */
-    /** 
-     * @param o
-     * @return <code>true</code> of the object <code>o</code> is a {@link FileResource} pointing to the same file as this resource. 
+
+    /**
+     * @param o the object to compare against this instance
+     * @return <code>true</code> of the object <code>o</code> is a {@link FileResource} pointing to the same file as this resource.
      */
     @Override
-    public boolean equals( Object o)
+    @SuppressWarnings("ReferenceEquality")
+    public boolean equals(Object o)
     {
         if (this == o)
             return true;
 
-        if (null == o || ! (o instanceof FileResource))
+        if (null == o || !(o instanceof FileResource))
             return false;
 
-        FileResource f=(FileResource)o;
+        FileResource f = (FileResource)o;
         return f._file == _file || (null != _file && _file.equals(f._file));
     }
 
-    /* ------------------------------------------------------------ */
     /**
      * @return the hashcode.
      */
     @Override
     public int hashCode()
     {
-       return null == _file ? super.hashCode() : _file.hashCode();
+        return null == _file ? super.hashCode() : _file.hashCode();
     }
-    
-    /* ------------------------------------------------------------ */
+
     @Override
     public void copyTo(File destination)
         throws IOException
     {
         if (isDirectory())
         {
-            IO.copyDir(getFile(),destination);
+            IO.copyDir(getFile(), destination);
         }
         else
         {
             if (destination.exists())
-                throw new IllegalArgumentException(destination+" exists");
-            IO.copy(getFile(),destination);
+                throw new IllegalArgumentException(destination + " exists");
+            IO.copy(getFile(), destination);
         }
     }
 
@@ -416,24 +467,23 @@ public class FileResource extends Resource
     {
         try
         {
-            return new URL(_uri);
+            return _uri.toURL();
         }
         catch (MalformedURLException e)
         {
             throw new IllegalStateException(e);
         }
     }
-    
+
     @Override
     public URI getURI()
     {
-        return _file.toURI();
+        return _uri;
     }
 
     @Override
     public String toString()
     {
-        return _uri;
+        return _uri.toString();
     }
-
 }
